#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <unistd.h>

#define BCM2708_PERI_BASE       0x20000000
#define REG_ACCESS(addr)        *(volatile unsigned*)((unsigned)peri_base_ptr + addr - 0x7e000000)
#define GPIO_DRIVE_REG          0x7e10002c

#define MAX_DRIVE               7
#define DRIVE_MASK              0x7
#define PASSWD                  0x5A
#define PASSWD_MASK             0xff000000UL

static const unsigned int drive_str[] = {2, 4, 6, 8, 10, 12, 14, 16};
static volatile unsigned *peri_base_ptr;

int mem_fd;
void *reg_map;

void setup_io()
{
   /* open /dev/mem */
   if ((mem_fd = open("/dev/mem", O_RDWR|O_SYNC) ) < 0) {
      printf("can't open /dev/mem \n");
      exit(-1);
   }

   /* mmap registers */
   reg_map = mmap(
      NULL,             //Any adddress in our space will do
      0x01000000,       //Map length
      PROT_READ|PROT_WRITE,// Enable reading & writting to mapped memory
      MAP_SHARED,       //Shared with other processes
      mem_fd,           //File to map
      BCM2708_PERI_BASE
   );

   close(mem_fd); //No need to keep mem_fd open after mmap

   if (reg_map == MAP_FAILED) {
      printf("mmap error %d\n", (int)reg_map);//errno also set!
      exit(-1);
   }

   // Always use volatile pointer!
   peri_base_ptr = (volatile unsigned *)reg_map;
}

int main(int argc, char **argv)
{
	unsigned old_reg, new_str;

	if ((argc != 2) || ((new_str = (unsigned)atoi(argv[1])) > MAX_DRIVE)) {
		printf("Usage: %s <0-7>\n", argv[0]);
		return -1;
	}

	setup_io();

	old_reg = REG_ACCESS(GPIO_DRIVE_REG);

        //printf("Register before write: 0x%.8X\n", old_reg);

	REG_ACCESS(GPIO_DRIVE_REG) = (old_reg & ~(PASSWD_MASK | DRIVE_MASK)) | (PASSWD << 24) | new_str;

	printf("GPIO strength changed from %umA to %umA\n", drive_str[old_reg & DRIVE_MASK], drive_str[new_str]);

	return 0;
}

